@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool/pool_process_query.c,v 1.13 2006/02/14 08:19:06 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool/pool_process_query.c,v 1.28 2006/06/22 08:17:16 y-asaba Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -39,6 +39,17 @@
 
 #include "pool.h"
 
+#define INIT_STATEMENT_LIST_SIZE 8
+
+/*
+ * prepared statement list
+ */
+typedef struct {
+	int size;
+	int cnt;
+	char **stmt_list;
+} PreparedStatementList;
+
 static POOL_STATUS NotificationResponse(POOL_CONNECTION *frontend, 
 										POOL_CONNECTION_POOL *backend);
 
@@ -107,6 +118,11 @@ static POOL_STATUS insert_lock(POOL_CONNECTION_POOL *backend, char *query);
 static char *get_insert_command_table_name(char *query);
 static char *skip_comment(char *query);
 
+static void add_prepared_list(PreparedStatementList *p, char *name);
+static void del_prepared_list(PreparedStatementList *p, char *name);
+static void reset_prepared_list(PreparedStatementList *p);
+static int send_deallocate(POOL_CONNECTION_POOL *backend, PreparedStatementList *p, int n);
+
 static POOL_CONNECTION_POOL_SLOT *slots[MAX_CONNECTION_SLOTS];
 
 static int in_load_balance;		/* non 0 if in load balance mode */
@@ -115,6 +131,10 @@ static int replication_was_enabled;		/* replication mode was enabled */
 static int master_slave_was_enabled;	/* master/slave mode was enabled */
 static int internal_transaction_started;		/* to issue table lock command a transaction
 												   has been started internally */
+static void (*pending_function)(PreparedStatementList *p, char *name) = NULL;
+static char *pending_prepared_name = NULL;
+
+static PreparedStatementList prepared_list; /* prepared statement name list */
 static int is_drop_database(char *query);		/* returns non 0 if this is a DROP DATABASE command */
 
 POOL_STATUS pool_process_query(POOL_CONNECTION *frontend, 
@@ -215,7 +235,7 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 			{
 				if (DUAL_MODE)
 					fds = select(Max(SECONDARY(backend)->fd,
-									 Max(frontend->fd, MASTER(backend)->fd)+1),
+									 Max(frontend->fd, MASTER(backend)->fd)) + 1,
 								 &readmask, &writemask, &exceptmask, NULL);
 				else
 					fds = select(Max(frontend->fd, MASTER(backend)->fd)+1,
@@ -263,7 +283,14 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 				if (status != POOL_CONTINUE)
 					return status;
 
-				continue;
+				if (kind != 0 || kind1 != 0)
+				{
+					pool_debug("kind(%02x) or kind1(%02x) != 0", kind, kind1);
+				}
+				else
+				{
+					continue;
+				}
 			}
 		}
 		else
@@ -311,7 +338,14 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 				if (status != POOL_CONTINUE)
 					return status;
 
-				continue;
+				if (kind != 0 || kind1 != 0)
+				{
+					pool_debug("cached kind(%02x) or kind1(%02x) != 0", kind, kind1);
+				}
+				else
+				{
+					continue;
+				}
 			}
 		}
 
@@ -513,6 +547,7 @@ static POOL_STATUS Query(POOL_CONNECTION *frontend,
 	char *string;
 	int len;
 	static char *sq = "show pool_status";
+	POOL_STATUS status;
 
 	if (query == NULL)	/* need to read query from frontend? */
 	{
@@ -536,7 +571,15 @@ static POOL_STATUS Query(POOL_CONNECTION *frontend,
 		string = query;
 	}
 
-	pool_debug("Query: %s", string);
+	/* log query to log file if neccessary */
+	if (pool_config.log_statement)
+	{
+		pool_log("statement: %s", string);
+	}
+	else
+	{
+		pool_debug("statement: %s", string);
+	}
 
 	/*
 	 * if this is DROP DATABASE command, send HUP signal to parent and
@@ -569,6 +612,37 @@ static POOL_STATUS Query(POOL_CONNECTION *frontend,
 		return POOL_CONTINUE;
 	}
 
+	if (frontend &&
+		(strncasecmp("prepare", string, 7) == 0 ||
+		strncasecmp("deallocate", string, 10) == 0))
+	{
+		char *query = string;
+		char *buf, *name;
+		query = skip_comment(query);
+
+		/* skip "prepare" or "deallocate" */
+		while (*query && !isspace(*query))
+			query++;
+
+		/* skip spaces */
+		while (*query && isspace(*query))
+			query++;
+
+		buf = strdup(query);
+		name = strtok(buf, "\t\r\n (;");
+		if (name && (*string == 'p' || *string == 'P'))
+		{
+			pending_function = add_prepared_list;
+			pending_prepared_name = strdup(name);
+		}
+		else if (name && (*string == 'd' || *string == 'D'))
+		{
+			pending_function = del_prepared_list;
+			pending_prepared_name = strdup(name);
+		}
+		free(buf);
+	}
+
 	/* load balance trick */
 	if (load_balance_enabled(backend, string))
 		start_load_balance(backend);
@@ -583,6 +657,7 @@ static POOL_STATUS Query(POOL_CONNECTION *frontend,
 	 * judge if we need to lock the table
 	 * to keep SERIAL data consistency among servers
 	 * conditions:
+	 * - replication is enabled
 	 * - protocol is V3
 	 * - statement is INSERT
 	 * - either "INSERT LOCK" comment exists or insert_lock directive specified
@@ -590,8 +665,9 @@ static POOL_STATUS Query(POOL_CONNECTION *frontend,
 	if (REPLICATION && need_insert_lock(backend, string))
 	{
 		/* start a transaction if needed and lock the table */
-		if (insert_lock(backend, string) != POOL_CONTINUE)
-			return POOL_END;
+		status = insert_lock(backend, string);
+		if (status != POOL_CONTINUE)
+			return status;
 	}
 
 	/* forward the query to the backend */
@@ -671,6 +747,8 @@ static POOL_STATUS Execute(POOL_CONNECTION *frontend,
 	int len;
 	int sendlen;
 	int i;
+	char kind;
+	int status;
 
 	/* read Execute packet */
 	if (pool_read(frontend, &len, sizeof(len)) < 0)
@@ -699,24 +777,38 @@ static POOL_STATUS Execute(POOL_CONNECTION *frontend,
 		sendlen = htonl(4);
 		if (pool_write_and_flush(cp, &sendlen, sizeof(sendlen)) < 0)
 		{
-			return POOL_END;
+			return POOL_ERROR;
 		}
 
 		if (!REPLICATION)
 			break;
-
-		/*
-		 * in "strict mode" we need to wait for backend completing the query.
-		 */
-		if (pool_config.replication_strict)
+		else if (pool_config.replication_strict)
 		{
-			pool_debug("waiting for backend[%d] completing the query", i);
+			pool_debug("waiting for backend completing the query");
 			if (synchronize(cp))
 				return POOL_END;
 		}
-
 	}
 
+	while ((kind = pool_read_kind(backend)),
+		   (kind != 'C' && kind != 'E' && kind != 'l' && kind != 's'))
+	{
+		if (kind < 0)
+		{
+			pool_error("Execute: pool_read_kind error");
+			return POOL_ERROR;
+		}
+
+		status = SimpleForwardToFrontend(kind, frontend, backend);
+		if (status != POOL_CONTINUE)
+			return status;
+		pool_flush(frontend);
+	}
+	status = SimpleForwardToFrontend(kind, frontend, backend);
+	if (status != POOL_CONTINUE)
+		return status;
+	pool_flush(frontend);
+
 	return POOL_CONTINUE;
 }
 
@@ -1630,7 +1722,7 @@ static POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 	if (pool_read(frontend, &fkind, 1) < 0)
 	{
 		pool_error("ProcessFrontendResponse: failed to read kind from frontend. fronend abnormally exited");
-		return POOL_END;
+		return POOL_ERROR;
 	}
 
 	pool_debug("read kind from frontend %c(%02x)", fkind, fkind);
@@ -1643,8 +1735,7 @@ static POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 				int len;
 				pool_read(frontend, &len, sizeof(len));
 			}
-			status = POOL_END;
-			break;
+			return POOL_END;
 
 		case 'Q':
 			status = Query(frontend, backend, NULL);
@@ -1655,7 +1746,6 @@ static POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 			status = Sync(frontend, backend);
 			break;
 */
-
 		case 'E':
 			status = Execute(frontend, backend);
 		break;
@@ -1687,6 +1777,8 @@ static POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 			break;
 	}
 
+	if (status != POOL_CONTINUE)
+		status = POOL_ERROR;
 	return status;
 }
 
@@ -2214,6 +2306,21 @@ POOL_STATUS SimpleForwardToFrontend(char kind, POOL_CONNECTION *frontend, POOL_C
 
 	pool_write(frontend, &kind, 1);
 
+	/*
+	 * Check if packet kind == 'C'(Command complete), '1'(Parse
+	 * complete), '3'(Close complete). If so, then register or unregister
+	 * pending prepared statement.
+	 */
+	if ((kind == 'C' || kind == '1' || kind == '3') &&
+		pending_function &&	pending_prepared_name)
+	{
+		pending_function(&prepared_list, pending_prepared_name);
+	}
+
+	free(pending_prepared_name);
+	pending_function = NULL;
+	pending_prepared_name = NULL;
+
 	status = pool_read(MASTER(backend), &len, sizeof(len));
 	if (status < 0)
 	{
@@ -2239,8 +2346,10 @@ POOL_STATUS SimpleForwardToFrontend(char kind, POOL_CONNECTION *frontend, POOL_C
 
 	pool_write(frontend, &len, sizeof(len));
 
-	len = ntohl(len);
-	len -= 4;
+	len = ntohl(len) - 4 ;
+
+	if (len <= 0)
+		return POOL_CONTINUE;
 
 	p = pool_read2(MASTER(backend), len);
 	if (p == NULL)
@@ -2248,13 +2357,120 @@ POOL_STATUS SimpleForwardToFrontend(char kind, POOL_CONNECTION *frontend, POOL_C
 
 	if (REPLICATION)
 	{
-		len1 = ntohl(len1);
-		len1 -= 4;
+		len1 = ntohl(len1) - 4;
+		if (len1 <= 0)
+			return POOL_CONTINUE;
+		
 		if (pool_read2(SECONDARY(backend), len1) == NULL)
 			return POOL_END;
 	}
 
-	return pool_write(frontend, p, len);
+	if (pool_write(frontend, p, len))
+		return POOL_END;
+
+	if (kind == 'E')		/* error response? */
+	{
+		int i, k;
+		int res1, res2;
+		char *p1;
+
+		/*
+		 * check if the error was PANIC or FATAL. If so, we just flush
+		 * the message and exit since the backend will close the
+		 * channel immediately.
+		 */
+		for (;;)
+		{
+			char e;
+
+			e = *p++;
+			if (e == '\0')
+				break;
+
+			if (e == 'S' && (strcasecmp("PANIC", p) == 0 || strcasecmp("FATAL", p) == 0))
+			{
+				pool_flush(frontend);
+				return POOL_END;
+			}
+			else
+			{
+				while (*p++)
+					;
+				continue;
+			}
+		}
+
+		for (i = 0;i < backend->num;i++)
+		{
+			POOL_CONNECTION *cp = backend->slots[i]->con;
+
+			/* We need to send "sync" message to backend in extend mode
+			 * so that it accepts next command.
+			 * Note that this may be overkill since client may send
+			 * it by itself. Moreover we do not need it in non-extend mode.
+			 * At this point we regard it is not harmfull since error resonse
+			 * will not be sent too frequently.
+			 */
+			pool_write(cp, "S", 1);
+			res1 = htonl(4);
+			if (pool_write_and_flush(cp, &res1, sizeof(res1)) < 0)
+			{
+				return POOL_END;
+			}
+
+			if (!REPLICATION)
+				break;
+		}
+
+		while ((k = pool_read_kind(backend)) != 'Z')
+		{
+			POOL_STATUS ret;
+			if (k < 0)
+			{
+				pool_error("SimpleForwardToBackend: pool_read_kind error");
+				return POOL_ERROR;
+			}
+
+			ret = SimpleForwardToFrontend(k, frontend, backend);
+			if (ret != POOL_CONTINUE)
+				return ret;
+			pool_flush(frontend);
+		}
+
+		status = pool_read(MASTER(backend), &res1, sizeof(res1));
+		if (status < 0)
+		{
+			pool_error("SimpleForwardToFrontend: error while reading message length");
+			return POOL_END;
+		}
+		res1 = ntohl(res1) - sizeof(res1);
+		p1 = pool_read2(MASTER(backend), res1);
+		if (p1 == NULL)
+			return POOL_END;
+	
+		if (REPLICATION)
+		{
+			status = pool_read(SECONDARY(backend), &res2, sizeof(res2));
+			if (status < 0)
+			{
+				pool_error("SimpleForwardToFrontend: error while reading message length from secondary backend");
+				return POOL_END;
+			}
+
+			res2 = ntohl(res2) - sizeof(res2);
+			p1 = pool_read2(SECONDARY(backend), res2);
+			if (p1 == NULL)
+				return POOL_END;
+
+			if (res1 != res2)
+			{
+				pool_debug("SimpleForwardToFrontend: length does not match between backends master(%d) secondary(%d) kind:(%c)",
+						   ntohl(res1), ntohl(res2), k);
+			}
+		}
+	}
+
+	return POOL_CONTINUE;
 }
 
 POOL_STATUS SimpleForwardToBackend(char kind, POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend)
@@ -2262,6 +2478,8 @@ POOL_STATUS SimpleForwardToBackend(char kind, POOL_CONNECTION *frontend, POOL_CO
 	int len;
 	int sendlen;
 	char *p;
+	int name_len;
+	char *name;
 
 	if (pool_write(MASTER(backend), &kind, 1))
 		return POOL_END;
@@ -2274,23 +2492,98 @@ POOL_STATUS SimpleForwardToBackend(char kind, POOL_CONNECTION *frontend, POOL_CO
 		return POOL_END;
 	}
 
+	if (pool_write(MASTER(backend), &sendlen, sizeof(sendlen)))
+		return POOL_END;
+	if (REPLICATION)
+		if (pool_write(SECONDARY(backend), &sendlen, sizeof(sendlen)))
+			return POOL_END;
+
 	len = ntohl(sendlen) - 4;
 
+	if (len <= 0)
+		return POOL_CONTINUE;
+
 	p = pool_read2(frontend, len);
 	if (p == NULL)
 		return POOL_END;
 
-	if (pool_write(MASTER(backend), &sendlen, sizeof(sendlen)))
+	if (pool_write_and_flush(MASTER(backend), p, len))
 		return POOL_END;
-	if (pool_write(MASTER(backend), p, len))
-		return POOL_END;
-
 	if (REPLICATION)
+		if (pool_write_and_flush(SECONDARY(backend), p, len))
+			return POOL_END;
+
+	if (kind == 'P' && *p)
 	{
-		if (pool_write(SECONDARY(backend), &sendlen, sizeof(sendlen)))
+		name_len = strlen(p) + 3;
+		name = malloc(name_len);
+		if (name == NULL)
+		{
+			pool_error("SimpleForwardToBackend: malloc failed: %s", strerror(errno));
 			return POOL_END;
-		if (pool_write(SECONDARY(backend), p, len))
+		}
+		sprintf(name, "\"%s\"", p);
+		pending_function = add_prepared_list;
+		pending_prepared_name = name;
+	}
+	else if (kind == 'C' && *p == 'S' && *(p + 1))
+	{
+		name_len = strlen(p + 1) + 3;
+		name = malloc(name_len);
+		if (name == NULL)
+		{
+			pool_error("SimpleForwardToBackend: malloc failed: %s", strerror(errno));
 			return POOL_END;
+		}
+		sprintf(name, "\"%s\"", p + 1);
+		pending_function = del_prepared_list;
+		pending_prepared_name = name;
+	}
+
+	if (kind == 'P' || kind == 'B' || kind == 'D' || kind == 'C')
+	{
+		int i;
+		int kind1;
+
+		for (i = 0;i < backend->num;i++)
+		{
+			POOL_CONNECTION *cp = backend->slots[i]->con;
+
+			/*
+			 * send "Flush" message so that backend notices us
+			 * the completion of the command
+			 */
+			pool_write(cp, "H", 1);
+			sendlen = htonl(4);
+			if (pool_write_and_flush(cp, &sendlen, sizeof(sendlen)) < 0)
+			{
+				return POOL_END;
+			}
+
+			if (!REPLICATION)
+				break;
+		}
+
+		if (kind == 'D' && *p == 'S')
+		{
+			kind1 = pool_read_kind(backend);
+			if (kind1 < 0)
+			{
+				pool_error("SimpleForwardToBackend: pool_read_kind error");
+				return POOL_ERROR;
+			}
+			SimpleForwardToFrontend(kind1, frontend, backend);
+			pool_flush(frontend);
+		}
+
+		kind1 = pool_read_kind(backend);
+		if (kind1 < 0)
+		{
+			pool_error("SimpleForwardToBackend: pool_read_kind error");
+			return POOL_ERROR;
+		}
+		SimpleForwardToFrontend(kind1, frontend, backend);
+		return pool_flush(frontend);
 	}
 
 	return POOL_CONTINUE;
@@ -2364,7 +2657,16 @@ static int reset_backend(POOL_CONNECTION_POOL *backend, int qcnt)
 	qn = pool_config.num_reset_queries;
 
 	if (qcnt >= qn)
-		return 2;
+	{
+		if (qcnt >= qn + prepared_list.cnt)
+		{
+			reset_prepared_list(&prepared_list);
+			return 2;
+		}
+
+		send_deallocate(backend, &prepared_list, qcnt - qn);
+		return 1;
+	}
 
 	query = pool_config.reset_query_list[qcnt];
 
@@ -2530,7 +2832,7 @@ void pool_send_error_message(POOL_CONNECTION *frontend, int protoMajor,
 
 		/* stop null */
 		len++;
-		*(data + len) = '\0';
+		*(data + len - 1) = '\0';
 
 		sendlen = len;
 		len = htonl(len + 4);
@@ -2584,16 +2886,12 @@ static POOL_STATUS do_command(POOL_CONNECTION *backend, char *query, int protoMa
 		return POOL_END;
 	}
 
-#ifdef NOT_USED
 	if (kind != 'C')
 	{
 		pool_error("do_command: backend does not successfully complete command %s status %c", query, kind);
 
-		/* the response must be an ERROR. handle it */
-		
-		return POOL_END;
 	}
-#endif
+
 	/*
 	 * read command tag of CommandComplete response
 	 */
@@ -2686,10 +2984,23 @@ static POOL_STATUS insert_lock(POOL_CONNECTION_POOL *backend, char *query)
 {
 	char *table;
 	char qbuf[1024];
+	POOL_STATUS status;
 	int i;
 
+	/* insert_lock can be used in V3 only */
 	if (MAJOR(backend) != PROTO_MAJOR_V3)
-		return 0;
+		return POOL_CONTINUE;
+
+	/* get table name */
+	table = get_insert_command_table_name(query);
+
+	/* could not get table name. probably wrong SQL command */
+	if (table == NULL)
+	{
+		return POOL_CONTINUE;
+	}
+
+	snprintf(qbuf, sizeof(qbuf), "LOCK TABLE %s IN SHARE ROW EXCLUSIVE MODE", table);
 
 	/* if we are not in a transaction block,
 	 * start a new transaction
@@ -2706,20 +3017,15 @@ static POOL_STATUS insert_lock(POOL_CONNECTION_POOL *backend, char *query)
 		internal_transaction_started = 1;
 	}
 
-	/* issue lock table command */
-	table = get_insert_command_table_name(query);
-	snprintf(qbuf, sizeof(qbuf), "LOCK TABLE %s IN SHARE ROW EXCLUSIVE MODE", table);
+	status = POOL_CONTINUE;
 
+	/* issue lock table command */
 	for (i = 0;i < backend->num;i++)
 	{
-		if (do_command(backend->slots[i]->con, qbuf, PROTO_MAJOR_V3, 0) != POOL_CONTINUE)
-		{
-			internal_transaction_started = 0;
-			return POOL_END;
-		}
+		status = do_command(backend->slots[i]->con, qbuf, PROTO_MAJOR_V3, 0);
 	}
 
-	return POOL_CONTINUE;
+	return status;
 }
 
 /*
@@ -2759,9 +3065,16 @@ static char *get_insert_command_table_name(char *query)
 	while (*query && isspace(*query))
 		query++;
 
-	/* get table */
+	/* get table name */
 	qbuf = strdup(query);
-	token = strtok(qbuf, " (");
+ 	token = strtok(qbuf, "\r\n\t (");
+
+	if (token == NULL)
+	{
+		pool_error("get_insert_command_table_name: could not get table name");
+		return NULL;
+	}
+
 	strncpy(table, token, sizeof(table));
 	free(qbuf);
 
@@ -2820,3 +3133,89 @@ static char *skip_comment(char *query)
 	}
 	return query;
 }
+
+void init_prepared_list(void)
+{
+	prepared_list.cnt = 0;
+	prepared_list.size = INIT_STATEMENT_LIST_SIZE;
+	prepared_list.stmt_list = malloc(sizeof(char *) * prepared_list.size);
+	if (prepared_list.stmt_list == NULL)
+	{
+		pool_error("init_prepared_list: malloc failed: %s", strerror(errno));
+		exit(1);
+	}
+}
+
+static void add_prepared_list(PreparedStatementList *p, char *name)
+{
+	if (p->cnt == p->size)
+	{
+		p->size *= 2;
+		p->stmt_list = realloc(p->stmt_list, sizeof(char *) * p->size);
+		if (p->stmt_list == NULL)
+		{
+			pool_error("add_prepared_list: realloc failed: %s", strerror(errno));
+			exit(1);
+		}
+	}
+
+	p->stmt_list[p->cnt++] = strdup(name);
+}
+
+static void del_prepared_list(PreparedStatementList *p, char *name)
+{
+	int i;
+
+	for (i = 0; i < p->cnt; i++)
+	{
+		if (strcmp(p->stmt_list[i], name) == 0)
+			break;
+	}
+	
+	if (i == p->cnt)
+		return;
+
+	free(p->stmt_list[i]);
+	if (i != p->cnt - 1)
+	{
+		memmove(&p->stmt_list[i], &p->stmt_list[i+1],
+				sizeof(char *) * (p->cnt - i - 1));
+	}
+	p->cnt--;
+}
+
+static void reset_prepared_list(PreparedStatementList *p)
+{
+	int i;
+
+	for (i = 0; i < p->cnt; i++)
+		free(p->stmt_list[i]);
+	p->cnt = 0;
+}
+
+static int send_deallocate(POOL_CONNECTION_POOL *backend, PreparedStatementList *p,
+					int n)
+{
+	char *query;
+	int len;
+
+	if (p->cnt <= n)
+		return 1;
+	
+	len = strlen(p->stmt_list[n]) + 12; /* "DEALLOCATE " + '\0' */
+	query = malloc(len);
+	if (query == NULL)
+	{
+		pool_error("send_deallocate: malloc failed: %s", strerror(errno));
+		exit(1);
+	}
+	sprintf(query, "DEALLOCATE %s", p->stmt_list[n]);
+	if (Query(NULL, backend, query) != POOL_CONTINUE)
+	{
+		free(query);
+		return 1;
+	}
+	free(query);
+
+	return 0;
+}
